@@ -22,22 +22,43 @@
 //! which is parsed using [`gimli`](https://github.com/gimli-rs/gimli).  The example CLI
 //! wrapper also uses symbol table information provided by the `object` crate.
 #![deny(missing_docs)]
+#![no_std]
+#![cfg_attr(not(feature = "std"), feature(alloc))]
+
+#[cfg(feature = "std")]
+#[macro_use]
+extern crate std;
+
+#[cfg(not(feature = "std"))]
+extern crate alloc;
+#[cfg(not(feature = "std"))]
+extern crate core as std;
 
 #[cfg(feature = "cpp_demangle")]
 extern crate cpp_demangle;
-extern crate fallible_iterator;
-extern crate gimli;
+pub extern crate fallible_iterator;
+pub extern crate gimli;
 extern crate intervaltree;
 extern crate lazycell;
-extern crate object;
+#[cfg(feature = "object")]
+pub extern crate object;
 #[cfg(feature = "rustc-demangle")]
 extern crate rustc_demangle;
 extern crate smallvec;
 
+#[cfg(feature = "std")]
+mod alloc {
+    pub use std::{borrow, rc, string, vec};
+}
+
+use alloc::borrow::Cow;
+#[cfg(feature = "object")]
+use alloc::rc::Rc;
+use alloc::string::{String, ToString};
+use alloc::vec::Vec;
+
 use std::cmp::Ordering;
-use std::borrow::Cow;
 use std::u64;
-use std::rc::Rc;
 
 use fallible_iterator::FallibleIterator;
 use intervaltree::{Element, IntervalTree};
@@ -50,21 +71,16 @@ struct Func<T> {
 }
 
 struct Lines<R: gimli::Reader> {
-    lnp: gimli::CompleteLineNumberProgram<R>,
-    sequences: Vec<gimli::LineNumberSequence<R>>,
+    lnp: gimli::CompleteLineProgram<R, R::Offset>,
+    sequences: Vec<gimli::LineSequence<R>>,
 }
 
 struct ResUnit<R>
 where
     R: gimli::Reader,
 {
-    dw_unit: gimli::CompilationUnitHeader<R, R::Offset>,
-    abbrevs: gimli::Abbreviations,
-    line_offset: gimli::DebugLineOffset<R::Offset>,
-    comp_dir: Option<R>,
-    comp_name: Option<R>,
+    dw_unit: gimli::Unit<R>,
     lang: Option<gimli::DwLang>,
-    base_addr: u64,
     lines: LazyCell<Result<Lines<R>, Error>>,
     funcs: LazyCell<Result<IntervalTree<u64, Func<R::Offset>>, Error>>,
 }
@@ -73,43 +89,13 @@ where
 ///
 /// Constructing a `Context` is somewhat costly, so users should aim to reuse `Context`s
 /// when performing lookups for many addresses in the same executable.
-pub struct Context<R>
+pub struct Context<R = gimli::EndianRcSlice<gimli::RunTimeEndian>>
 where
     R: gimli::Reader,
 {
     unit_ranges: Vec<(gimli::Range, usize)>,
     units: Vec<ResUnit<R>>,
-    sections: DebugSections<R>,
-}
-
-fn read_ranges<R: gimli::Reader>(
-    entry: &gimli::DebuggingInformationEntry<R, R::Offset>,
-    range_lists: &gimli::RangeLists<R>,
-    dw_unit: &gimli::CompilationUnitHeader<R, R::Offset>,
-    base_addr: u64,
-) -> Result<Option<WrapRangeIter<R>>, Error> {
-    Ok(Some(match entry.attr_value(gimli::DW_AT_ranges)? {
-        None => {
-            let low_pc = match entry.attr_value(gimli::DW_AT_low_pc)? {
-                Some(gimli::AttributeValue::Addr(low_pc)) => low_pc,
-                _ => return Ok(None), // neither ranges nor low_pc => None
-            };
-            let high_pc = match entry.attr_value(gimli::DW_AT_high_pc)? {
-                Some(gimli::AttributeValue::Addr(high_pc)) => high_pc,
-                Some(gimli::AttributeValue::Udata(x)) => low_pc + x,
-                _ => return Ok(None), // only low_pc, no high_pc? wtf is this? TODO: perhaps return error
-            };
-            WrapRangeIter::Synthetic(Some(gimli::Range {
-                begin: low_pc,
-                end: high_pc,
-            }))
-        }
-        Some(gimli::AttributeValue::RangeListsRef(rr)) => {
-            let ranges = range_lists.ranges(rr, dw_unit.version(), dw_unit.address_size(), base_addr)?;
-            WrapRangeIter::Real(ranges)
-        }
-        _ => unreachable!(),
-    }))
+    sections: gimli::Dwarf<R>,
 }
 
 impl Context<gimli::EndianRcSlice<gimli::RunTimeEndian>> {
@@ -121,16 +107,19 @@ impl Context<gimli::EndianRcSlice<gimli::RunTimeEndian>> {
     ///
     /// Performance sensitive applications may want to use `Context::from_sections`
     /// with a more specialised `gimli::Reader` implementation.
-    pub fn new<'input, 'data, O: object::Object<'input, 'data>>(file: &O) -> Result<Self, Error> {
+    #[cfg(feature = "object")]
+    pub fn new<'data, 'file, O: object::Object<'data, 'file>>(
+        file: &'file O,
+    ) -> Result<Self, Error> {
         let endian = if file.is_little_endian() {
             gimli::RunTimeEndian::Little
         } else {
             gimli::RunTimeEndian::Big
         };
 
-        fn load_section<'input, 'data, O, S, Endian>(file: &O, endian: Endian) -> S
+        fn load_section<'data, 'file, O, S, Endian>(file: &'file O, endian: Endian) -> S
         where
-            O: object::Object<'input, 'data>,
+            O: object::Object<'data, 'file>,
             S: gimli::Section<gimli::EndianRcSlice<Endian>>,
             Endian: gimli::Endianity,
         {
@@ -139,19 +128,27 @@ impl Context<gimli::EndianRcSlice<gimli::RunTimeEndian>> {
         }
 
         let debug_abbrev: gimli::DebugAbbrev<_> = load_section(file, endian);
+        let debug_addr: gimli::DebugAddr<_> = load_section(file, endian);
         let debug_info: gimli::DebugInfo<_> = load_section(file, endian);
         let debug_line: gimli::DebugLine<_> = load_section(file, endian);
+        let debug_line_str: gimli::DebugLineStr<_> = load_section(file, endian);
         let debug_ranges: gimli::DebugRanges<_> = load_section(file, endian);
         let debug_rnglists: gimli::DebugRngLists<_> = load_section(file, endian);
         let debug_str: gimli::DebugStr<_> = load_section(file, endian);
+        let debug_str_offsets: gimli::DebugStrOffsets<_> = load_section(file, endian);
+        let default_section = gimli::EndianRcSlice::new(Rc::from(&[][..]), endian);
 
         Context::from_sections(
             debug_abbrev,
+            debug_addr,
             debug_info,
             debug_line,
+            debug_line_str,
             debug_ranges,
             debug_rnglists,
             debug_str,
+            debug_str_offsets,
+            default_section,
         )
     }
 }
@@ -160,73 +157,63 @@ impl<R: gimli::Reader> Context<R> {
     /// Construct a new `Context` from DWARF sections.
     pub fn from_sections(
         debug_abbrev: gimli::DebugAbbrev<R>,
+        debug_addr: gimli::DebugAddr<R>,
         debug_info: gimli::DebugInfo<R>,
         debug_line: gimli::DebugLine<R>,
+        debug_line_str: gimli::DebugLineStr<R>,
         debug_ranges: gimli::DebugRanges<R>,
         debug_rnglists: gimli::DebugRngLists<R>,
         debug_str: gimli::DebugStr<R>,
+        debug_str_offsets: gimli::DebugStrOffsets<R>,
+        default_section: R,
     ) -> Result<Self, Error> {
-        let range_lists = gimli::RangeLists::new(debug_ranges, debug_rnglists)?;
+        let sections = gimli::Dwarf {
+            debug_abbrev,
+            debug_addr,
+            debug_info,
+            debug_line,
+            debug_line_str,
+            debug_str,
+            debug_str_offsets,
+            debug_str_sup: default_section.clone().into(),
+            debug_types: default_section.clone().into(),
+            locations: gimli::LocationLists::new(default_section.clone().into(), default_section.clone().into()),
+            ranges: gimli::RangeLists::new(debug_ranges, debug_rnglists),
+        };
 
         let mut unit_ranges = Vec::new();
         let mut res_units = Vec::new();
-        let mut units = debug_info.units();
-        while let Some(dw_unit) = units.next()? {
+        let mut units = sections.units();
+        while let Some(header) = units.next()? {
             let unit_id = res_units.len();
+            let dw_unit = sections.unit(header)?;
 
-            let abbrevs = dw_unit.abbreviations(&debug_abbrev)?;
-
-            let dlr;
-            let dcd;
-            let dcn;
-            let base_addr;
             let lang;
             {
-                let mut cursor = dw_unit.entries(&abbrevs);
+                let mut cursor = dw_unit.entries();
 
                 let unit = match cursor.next_dfs()? {
                     Some((_, unit)) if unit.tag() == gimli::DW_TAG_compile_unit => unit,
                     _ => continue, // wtf?
                 };
 
-                dlr = match unit.attr_value(gimli::DW_AT_stmt_list)? {
-                    Some(gimli::AttributeValue::DebugLineRef(dlr)) => dlr,
-                    _ => unreachable!(),
-                };
-                dcd = unit.attr(gimli::DW_AT_comp_dir)?
-                    .and_then(|x| x.string_value(&debug_str));
-                dcn = unit.attr(gimli::DW_AT_name)?
-                    .and_then(|x| x.string_value(&debug_str));
-                base_addr = match unit.attr_value(gimli::DW_AT_low_pc)? {
-                    Some(gimli::AttributeValue::Addr(addr)) => addr,
-                    None => 0, // ThinLTO yields inline-only compilation units; this is valid
-                    _ => unreachable!(),
-                };
                 lang = match unit.attr_value(gimli::DW_AT_language)? {
                     Some(gimli::AttributeValue::Language(lang)) => Some(lang),
                     _ => None,
                 };
-                if let Some(mut ranges) =
-                    read_ranges(unit, &range_lists, &dw_unit, base_addr)?
-                {
-                    while let Some(range) = ranges.next()? {
-                        if range.begin == range.end {
-                            continue;
-                        }
-
-                        unit_ranges.push((range, unit_id));
+                let mut ranges = sections.unit_ranges(&dw_unit)?;
+                while let Some(range) = ranges.next()? {
+                    if range.begin == range.end {
+                        continue;
                     }
+
+                    unit_ranges.push((range, unit_id));
                 }
             }
 
             res_units.push(ResUnit {
                 dw_unit,
-                abbrevs,
-                line_offset: dlr,
-                comp_dir: dcd,
-                comp_name: dcn,
                 lang,
-                base_addr,
                 lines: LazyCell::new(),
                 funcs: LazyCell::new(),
             });
@@ -253,11 +240,7 @@ impl<R: gimli::Reader> Context<R> {
         Ok(Context {
             units: res_units,
             unit_ranges,
-            sections: DebugSections {
-                debug_str,
-                range_lists,
-                debug_line,
-            },
+            sections,
         })
     }
 }
@@ -266,60 +249,54 @@ impl<R> ResUnit<R>
 where
     R: gimli::Reader,
 {
-    fn parse_lines(&self, sections: &DebugSections<R>) -> Result<&Lines<R>, Error> {
+    fn parse_lines(&self) -> Result<Option<&Lines<R>>, Error> {
+        let ilnp = match self.dw_unit.line_program {
+            Some(ref ilnp) => ilnp,
+            None => return Ok(None),
+        };
         self.lines
             .borrow_with(|| {
-                let ilnp = sections.debug_line.program(
-                    self.line_offset,
-                    self.dw_unit.address_size(),
-                    self.comp_dir.clone(),
-                    self.comp_name.clone(),
-                )?;
-
-                let (lnp, mut sequences) = ilnp.sequences()?;
+                let (lnp, mut sequences) = ilnp.clone().sequences()?;
                 sequences.retain(|x| x.start != 0);
                 sequences.sort_by_key(|x| x.start);
                 Ok(Lines { lnp, sequences })
             })
             .as_ref()
+            .map(Some)
             .map_err(Error::clone)
     }
 
     fn parse_functions(
         &self,
-        sections: &DebugSections<R>,
+        sections: &gimli::Dwarf<R>,
     ) -> Result<&IntervalTree<u64, Func<R::Offset>>, Error> {
         self.funcs
             .borrow_with(|| {
                 let mut results = Vec::new();
                 let mut depth = 0;
-                let mut cursor = self.dw_unit.entries(&self.abbrevs);
+                let mut cursor = self.dw_unit.entries();
                 while let Some((d, entry)) = cursor.next_dfs()? {
                     depth += d;
                     match entry.tag() {
                         gimli::DW_TAG_subprogram | gimli::DW_TAG_inlined_subroutine => {
-                            // may be an inline-only function and thus not have any ranges
-                            if let Some(mut ranges) = read_ranges(
-                                entry,
-                                &sections.range_lists,
+                            let mut ranges = sections.die_ranges(
                                 &self.dw_unit,
-                                self.base_addr,
-                            )? {
-                                while let Some(range) = ranges.next()? {
-                                    // Ignore invalid DWARF so that a query of 0 does not give
-                                    // a long list of matches.
-                                    // TODO: don't ignore if there is a section at this address
-                                    if range.begin == 0 {
-                                        continue;
-                                    }
-                                    results.push(Element {
-                                        range: range.begin..range.end,
-                                        value: Func {
-                                            entry_off: entry.offset(),
-                                            depth,
-                                        },
-                                    });
+                                entry,
+                            )?;
+                            while let Some(range) = ranges.next()? {
+                                // Ignore invalid DWARF so that a query of 0 does not give
+                                // a long list of matches.
+                                // TODO: don't ignore if there is a section at this address
+                                if range.begin == 0 {
+                                    continue;
                                 }
+                                results.push(Element {
+                                    range: range.begin..range.end,
+                                    value: Func {
+                                        entry_off: entry.offset(),
+                                        depth,
+                                    },
+                                });
                             }
                         }
                         _ => (),
@@ -334,12 +311,6 @@ where
     }
 }
 
-struct DebugSections<R: gimli::Reader> {
-    debug_str: gimli::DebugStr<R>,
-    range_lists: gimli::RangeLists<R>,
-    debug_line: gimli::DebugLine<R>,
-}
-
 /// An iterator over function frames.
 pub struct FrameIter<'ctx, R>
 where
@@ -347,7 +318,7 @@ where
 {
     unit_id: usize,
     units: &'ctx Vec<ResUnit<R>>,
-    sections: &'ctx DebugSections<R>,
+    sections: &'ctx gimli::Dwarf<R>,
     funcs: smallvec::IntoIter<[&'ctx Func<R::Offset>; 16]>,
     next: Option<Location>,
 }
@@ -382,13 +353,14 @@ impl<R: gimli::Reader> FunctionName<R> {
 /// Demangle a symbol name using the demangling scheme for the given language.
 ///
 /// Returns `None` if demangling failed or is not required.
+#[allow(unused_variables)]
 pub fn demangle(name: &str, language: gimli::DwLang) -> Option<String> {
     match language {
         #[cfg(feature = "rustc-demangle")]
         gimli::DW_LANG_Rust => rustc_demangle::try_demangle(name)
             .ok()
             .as_ref()
-            .map(ToString::to_string),
+            .map(|x| format!("{:#}", x)),
         #[cfg(feature = "cpp_demangle")]
         gimli::DW_LANG_C_plus_plus
         | gimli::DW_LANG_C_plus_plus_03
@@ -417,23 +389,6 @@ pub fn demangle_auto(name: Cow<str>, language: Option<gimli::DwLang>) -> Cow<str
     }.map(Cow::from).unwrap_or(name)
 }
 
-enum WrapRangeIter<R: gimli::Reader> {
-    Real(gimli::RngListIter<R>),
-    Synthetic(Option<gimli::Range>),
-}
-
-impl<R: gimli::Reader> FallibleIterator for WrapRangeIter<R> {
-    type Item = gimli::Range;
-    type Error = gimli::Error;
-
-    fn next(&mut self) -> Result<Option<gimli::Range>, gimli::Error> {
-        match *self {
-            WrapRangeIter::Real(ref mut ri) => ri.next(),
-            WrapRangeIter::Synthetic(ref mut range) => Ok(range.take()),
-        }
-    }
-}
-
 /// A source location.
 pub struct Location {
     /// The file name.
@@ -515,9 +470,12 @@ where
     fn find_location(
         &self,
         probe: u64,
-        sections: &DebugSections<R>,
+        sections: &gimli::Dwarf<R>,
     ) -> Result<Option<Location>, Error> {
-        let lines = self.parse_lines(sections)?;
+        let lines = match self.parse_lines()? {
+            Some(lines) => lines,
+            None => return Ok(None),
+        };
         let idx = lines.sequences.binary_search_by(|ln| {
             if probe < ln.start {
                 Ordering::Greater
@@ -550,7 +508,7 @@ where
         }
 
         let file = match file {
-            Some(file) => Some(self.render_file(file, lines)?),
+            Some(file) => Some(self.render_file(file, lines, sections)?),
             None => None,
         };
 
@@ -559,20 +517,21 @@ where
 
     fn render_file(
         &self,
-        file: &gimli::FileEntry<R>,
+        file: &gimli::FileEntry<R, R::Offset>,
         lines: &Lines<R>,
+        sections: &gimli::Dwarf<R>,
     ) -> Result<String, gimli::Error> {
-        let mut path = if let Some(ref comp_dir) = self.comp_dir {
+        let mut path = if let Some(ref comp_dir) = self.dw_unit.comp_dir {
             String::from(comp_dir.to_string_lossy()?.as_ref())
         } else {
             String::new()
         };
 
         if let Some(directory) = file.directory(lines.lnp.header()) {
-            path_push(&mut path, directory.to_string_lossy()?.as_ref());
+            path_push(&mut path, sections.attr_string(&self.dw_unit, directory)?.to_string_lossy()?.as_ref());
         }
 
-        path_push(&mut path, file.path_name().to_string_lossy()?.as_ref());
+        path_push(&mut path, sections.attr_string(&self.dw_unit, file.path_name())?.to_string_lossy()?.as_ref());
 
         Ok(path)
     }
@@ -594,7 +553,7 @@ type Error = gimli::Error;
 fn name_attr<'abbrev, 'unit, R>(
     entry: &gimli::DebuggingInformationEntry<'abbrev, 'unit, R, R::Offset>,
     unit: &ResUnit<R>,
-    sections: &DebugSections<R>,
+    sections: &gimli::Dwarf<R>,
     units: &[ResUnit<R>],
     recursion_limit: usize,
 ) -> Result<Option<R>, Error>
@@ -605,18 +564,18 @@ where
         return Ok(None);
     }
 
-    if let Some(attr) = entry.attr(gimli::DW_AT_linkage_name)? {
-        if let Some(val) = attr.string_value(&sections.debug_str) {
+    if let Some(attr) = entry.attr_value(gimli::DW_AT_linkage_name)? {
+        if let Ok(val) = sections.attr_string(&unit.dw_unit, attr) {
             return Ok(Some(val));
         }
     }
-    if let Some(attr) = entry.attr(gimli::DW_AT_MIPS_linkage_name)? {
-        if let Some(val) = attr.string_value(&sections.debug_str) {
+    if let Some(attr) = entry.attr_value(gimli::DW_AT_MIPS_linkage_name)? {
+        if let Ok(val) = sections.attr_string(&unit.dw_unit, attr) {
             return Ok(Some(val));
         }
     }
-    if let Some(attr) = entry.attr(gimli::DW_AT_name)? {
-        if let Some(val) = attr.string_value(&sections.debug_str) {
+    if let Some(attr) = entry.attr_value(gimli::DW_AT_name)? {
+        if let Ok(val) = sections.attr_string(&unit.dw_unit, attr) {
             return Ok(Some(val));
         }
     }
@@ -626,7 +585,7 @@ where
         .or(entry.attr_value(gimli::DW_AT_specification)?);
     match next {
         Some(gimli::AttributeValue::UnitRef(offset)) => {
-            let mut entries = unit.dw_unit.entries_at_offset(&unit.abbrevs, offset)?;
+            let mut entries = unit.dw_unit.entries_at_offset(offset)?;
             if let Some((_, entry)) = entries.next_dfs()? {
                 return name_attr(entry, unit, sections, units, recursion_limit - 1);
             } else {
@@ -635,10 +594,14 @@ where
         }
         Some(gimli::AttributeValue::DebugInfoRef(dr)) => if let Some((unit, offset)) = units
             .iter()
-            .filter_map(|unit| dr.to_unit_offset(&unit.dw_unit).map(|uo| (unit, uo)))
+            .filter_map(|unit| {
+                gimli::UnitSectionOffset::DebugInfoOffset(dr)
+                    .to_unit_offset(&unit.dw_unit)
+                    .map(|uo| (unit, uo))
+            })
             .next()
         {
-            let mut entries = unit.dw_unit.entries_at_offset(&unit.abbrevs, offset)?;
+            let mut entries = unit.dw_unit.entries_at_offset(offset)?;
             if let Some((_, entry)) = entries.next_dfs()? {
                 return name_attr(entry, unit, sections, units, recursion_limit - 1);
             }
@@ -670,8 +633,7 @@ where
 
         let unit = &self.units[self.unit_id];
 
-        let mut cursor = unit.dw_unit
-            .entries_at_offset(&unit.abbrevs, func.entry_off)?;
+        let mut cursor = unit.dw_unit.entries_at_offset(func.entry_off)?;
         let (_, entry) = cursor
             .next_dfs()?
             .expect("DIE we read a while ago is no longer readable??");
@@ -682,9 +644,13 @@ where
         if entry.tag() == gimli::DW_TAG_inlined_subroutine {
             let file = match entry.attr_value(gimli::DW_AT_call_file)? {
                 Some(gimli::AttributeValue::FileIndex(fi)) => {
-                    let lines = unit.parse_lines(self.sections)?;
-                    match lines.lnp.header().file(fi) {
-                        Some(file) => Some(unit.render_file(file, lines)?),
+                    match unit.parse_lines()? {
+                        Some(lines) => {
+                            match lines.lnp.header().file(fi) {
+                                Some(file) => Some(unit.render_file(file, lines, self.sections)?),
+                                None => None,
+                            }
+                        }
                         None => None,
                     }
                 }
